/*
 * lxlist-p.n.c
 * Copyright (C) 2011-2017, 2019, 2022, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * there'd be a chance that the list object gets destroyed while evaluation
 * the element function.  and as the object is still accessed after the many
 * evaluations, that'd be trouble.  though I guess only badly written C
 * functions could destroy the list object.
 */

#include <stddef.h>

#include <e4.h>
#include <l_list.h>
#include <lxcall.h>
#include <lxcast.h>
#include <lxlist-close.h>
#include <lxlist-drive.h>
#include <lxlist-inter.h>
#include <lxlist-types.h>

#define __LINK_COPY_DATA__		0

#define lxlist(list) \
    ((struct lxlist_type *) (list))

typedef struct port_type {
    int (*call)(void *, int, void *, struct x1f4_dispatch_type **),
	(*function)(void *, void *, void **),
	(*jack)(struct port_type *, struct fxdata_type *,
		struct x1f4_dxcast_type *, int *), link, type;
    struct cxroll_type cxroll;
    struct dxlist_type *dxlist_data;
    struct excase_type *excase_data;
    struct lxlist_type *lxlist_data;
    struct lxtext_type *lxtext_data, *lxtext_lift;
    struct x1f4_dispatch_type *dispatch_data;
    unsigned count, size;
    void *back, *context, *output, *side;
} port_type;

typedef struct rule_type {
    struct direct_type *direct_data;
} rule_type;

extern const void *const x1f4_a1_walk_link;

static int copy_data(struct port_type *, struct fxdata_type *,
		     struct x1f4_dxcast_type *, int *);
static int ever_data(struct port_type *, struct fxdata_type *,
		     struct x1f4_dxcast_type *, int *);
static int fast_data(struct port_type *, struct fxdata_type *,
		     struct x1f4_dxcast_type *, int *);
#if __LINK_COPY_DATA__
static int free_data(struct port_type *, struct fxdata_type *, int, int);
#endif				/* __LINK_COPY_DATA__ */
static int jack_data(struct port_type *, struct fxdata_type *,
		     struct x1f4_dxcast_type *, int, int *);
static int less_data(void *, int, void *, struct x1f4_dispatch_type **);
static int miss_data(void *, int, void *, struct x1f4_dispatch_type **);
static int poll_data(void *, void *, void **);
static int roll_data(void *, void *, void **);
static int rule_list(void *, void *);
static int slip_port(void *);
static int spin_port(struct port_type *);
static int swap_port(void *);
static int wind_list(struct port_type *, void *);

static int
copy_data(struct port_type *port_data, struct fxdata_type *fxdata_data,
	  struct x1f4_dxcast_type *dxcast_data, int *jack)
{
    int status;
    struct lxtext_type *lxtext_data;

    *jack = port_data->link;

    lxtext_data = port_data->lxtext_lift;

    status = lxtext_data->lxtype.copy
	(lxtext_data->lxtype.context, NULL, (void *) fxdata_data,
	 I_USER(&dxcast_data->data), port_data->excase_data);
    if (status) {
	*jack = ~0;
    } else {
#if __LINK_COPY_DATA__
#else
	(*port_data->cxroll.miss)++;
#endif				/* __LINK_COPY_DATA__ */
    }

    return status;
}


static int
ever_data(struct port_type *port_data, struct fxdata_type *fxdata_data,
	  struct x1f4_dxcast_type *dxcast_data, int *jack)
{
    *jack = ~0;

#if __LINK_COPY_DATA__
#else
    (*port_data->cxroll.miss)++;
#endif				/* __LINK_COPY_DATA__ */

    *fxdata_data = dxcast_data->data;

    return 0;
}


static int
fast_data(struct port_type *port_data, struct fxdata_type *fxdata_data,
	  struct x1f4_dxcast_type *dxcast_data, int *jack)
{
    int status;
    struct lxtext_type *lxtext_data;

    *jack = port_data->link;

    lxtext_data = port_data->lxtext_lift;

    status = lxtext_data->lxtype.node
	(lxtext_data->lxtype.context, (void *) fxdata_data,
	 I_USER(&dxcast_data->data), port_data->excase_data);
    if (status) {
	*jack = ~0;
    } else {
#if __LINK_COPY_DATA__
#else
	(*port_data->cxroll.miss)++;
#endif				/* __LINK_COPY_DATA__ */
    }

    return status;
}


#if __LINK_COPY_DATA__
static int
free_data(struct port_type *port_data, struct fxdata_type *fxdata_data,
	  int sink, int status)
{
    int excess, (*free)(void *, void *, struct excase_type **,
			const struct excase_type *);
    struct lxtext_type *lxtext_slip;

    lxtext_slip = port_data->lxtext_data + sink;

    free = lxtext_slip->lxtype.free;
    if (free) {
	excess = free
	    (lxtext_slip->lxtype.context, I_USER(fxdata_data), NULL,
	     port_data->excase_data);
	if (excess) {
	    status = _libx1f4i0_lxcall_land_slip(excess, status);
	}
    }

    return status;
}
#endif				/* __LINK_COPY_DATA__ */


static int
jack_data(struct port_type *port_data, struct fxdata_type *fxdata_data,
	  struct x1f4_dxcast_type *dxcast_data, int type, int *jack)
{
    int status;
    struct lxtext_type *lxtext_data;
    unsigned miss;

    lxtext_data = port_data->lxtext_data;

    close_frame(miss, lxtext_data);

    if (miss) {
	*jack = lxtext_data - port_data->lxtext_data;

	if (lxtext_data->lxtype.flags & X1f4_LX_LINK_ACCESS) {
	    status = lxtext_data->lxtype.node
		(lxtext_data->lxtype.context, (void *) fxdata_data,
		 I_USER(&dxcast_data->data), port_data->excase_data);
	    if (status) {
		*jack = ~0;
	    } else {
#if __LINK_COPY_DATA__
#else
		(*port_data->cxroll.miss)++;
#endif				/* __LINK_COPY_DATA__ */
	    }
	} else {
	    int (*copy)(void *, void *, void **, const void *,
			const struct x1f4_nodelink_type *);

	    copy = lxtext_data->lxtype.copy;
	    if (copy) {
		status = copy
		    (lxtext_data->lxtype.context, NULL, (void *) fxdata_data,
		     I_USER(&dxcast_data->data), port_data->excase_data);
		if (status) {
		    *jack = ~0;
		} else {
#if __LINK_COPY_DATA__
#else
		    (*port_data->cxroll.miss)++;
#endif				/* __LINK_COPY_DATA__ */
		}
	    } else {
		status = 0;

#if __LINK_COPY_DATA__
#else
		(*port_data->cxroll.miss)++;
#endif				/* __LINK_COPY_DATA__ */

		*fxdata_data = dxcast_data->data;
	    }
	}
    } else {
	*jack = ~0;

	status = 0;

#if __LINK_COPY_DATA__
#else
	(*port_data->cxroll.miss)++;
#endif				/* __LINK_COPY_DATA__ */

	*fxdata_data = dxcast_data->data;
    }

    return status;
}


static int
less_data(void *text, int status, void *output,
	  struct x1f4_dispatch_type **dispatch)
{
    int recall = 1;
    struct port_type *port_data;

    port_data = text;

    if (status) {
	spin_port(port_data);
    } else {
	do {
	    int link;
#if __LINK_COPY_DATA__
	    int sink;
#endif				/* __LINK_COPY_DATA__ */
#if __LINK_COPY_DATA__
	    struct fxdata_type fxdata;
#endif				/* __LINK_COPY_DATA__ */
	    struct direct_type *direct_data;

	    direct_data = port_data->cxroll.direct_data;

#if __LINK_COPY_DATA__
	    sink = direct_data->link;
#endif				/* __LINK_COPY_DATA__ */

#if __LINK_COPY_DATA__
	    fxdata = direct_data->fxdata;
#endif				/* __LINK_COPY_DATA__ */

	    status = port_data->jack
		(port_data, &direct_data->fxdata, I_USER(output), &link);

#if __LINK_COPY_DATA__
	    if (~sink) {
		status = free_data(port_data, &fxdata, sink, status);
	    }
#endif				/* __LINK_COPY_DATA__ */

	    port_data->count--;

	    direct_data->link = link;
	    direct_data->type = port_data->type;

	    port_data->cxroll.direct_data = direct_data + 1;

	    if (status) {
		break;
	    }

	    /*
	     * call this (_lift_call) after data swapping, as it has more to
	     * do with the next item.
	     */
	    recall = 0;

	    status = port_data->call(port_data->back, 0, output, dispatch);
	    if (status) {
                break;
	    }

	    if (port_data->count) {
		struct x1f4_dispatch_type *dispatch_data;

		dispatch_data = port_data->dispatch_data;

		dispatch_data->back = port_data;
		dispatch_data->call = less_data;
	    } else {
		status = wind_list(port_data, output);

		/*
		 * *dispatch set NULL by _lift_call
		 */
	    }
	} while (0);
    }

    if (recall) {
	status = port_data->call(port_data->back, status, output, dispatch);
    }

    return status;
}


static int
miss_data(void *text, int status, void *output,
	  struct x1f4_dispatch_type **dispatch)
{
    int recall = 1;
    struct port_type *port_data;

    port_data = text;

    if (status) {
	spin_port(port_data);
    } else {
	int type;
	struct x1f4_dxcast_type *dxcast_data;

	dxcast_data = I_USER(output);

	type = dxcast_data->type;

	do {
	    int link;
#if __LINK_COPY_DATA__
	    int sink;
#endif				/* __LINK_COPY_DATA__ */
#if __LINK_COPY_DATA__
	    struct fxdata_type fxdata;
#endif				/* __LINK_COPY_DATA__ */
	    struct direct_type *direct_data;

	    if (type == X1f4_E4_VOID) {
		status = _libx1f4i0_lxlist_stat_gulf(port_data->lxlist_data);
		break;
	    }

	    direct_data = port_data->cxroll.direct_data;

#if __LINK_COPY_DATA__
	    sink = direct_data->link;
#endif				/* __LINK_COPY_DATA__ */

#if __LINK_COPY_DATA__
	    fxdata = direct_data->fxdata;
#endif				/* __LINK_COPY_DATA__ */

	    status = jack_data
		(port_data, &direct_data->fxdata, dxcast_data, type, &link);

#if __LINK_COPY_DATA__
	    if (~sink) {
		status = free_data(port_data, &fxdata, sink, status);
	    }
#endif				/* __LINK_COPY_DATA__ */

	    port_data->count--;

	    direct_data->link = link;
	    direct_data->type = type;

	    port_data->cxroll.direct_data = direct_data + 1;

	    if (status) {
		break;
	    }

	    /*
	     * call this (_lift_call) after data swapping, as it has more to
	     * do with the next item.
	     */
	    recall = 0;

	    status = port_data->call(port_data->back, 0, output, dispatch);
	    if (status) {
                break;
	    }

	    if (port_data->count) {
		struct x1f4_dispatch_type *dispatch_data;

		dispatch_data = port_data->dispatch_data;

		dispatch_data->back = port_data;
		dispatch_data->call = miss_data;
	    } else {
		status = wind_list(port_data, output);

		/*
		 * *dispatch set NULL by _lift_call
		 */
	    }
	} while (0);
    }

    if (recall) {
	status = port_data->call(port_data->back, status, output, dispatch);
    }

    return status;
}


static int
poll_data(void *context, void *output, void **input)
{
    int status;
    struct port_type *port_data;

    port_data = context;

    status = port_data->function(port_data->context, output, input);
    if (status) {
    } else {
	do {
	    int link;
#if __LINK_COPY_DATA__
	    int sink;
#endif				/* __LINK_COPY_DATA__ */
#if __LINK_COPY_DATA__
	    struct fxdata_type fxdata;
#endif				/* __LINK_COPY_DATA__ */
	    struct direct_type *direct_data;

	    direct_data = port_data->cxroll.direct_data;

#if __LINK_COPY_DATA__
	    sink = direct_data->link;
#endif				/* __LINK_COPY_DATA__ */

#if __LINK_COPY_DATA__
	    fxdata = direct_data->fxdata;
#endif				/* __LINK_COPY_DATA__ */

	    status = port_data->jack
		(port_data, &direct_data->fxdata,
		 (void *)
		 ((char *) output - offsetof(struct x1f4_dxcast_type, data)),
		 &link);

#if __LINK_COPY_DATA__
	    if (~sink) {
		status = free_data(port_data, &fxdata, sink, status);
	    }
#endif				/* __LINK_COPY_DATA__ */

	    port_data->count--;

	    direct_data->link = link;
	    direct_data->type = port_data->type;

	    port_data->cxroll.direct_data = direct_data + 1;

	    if (status) {
		break;
	    }
	    if (port_data->count) {
	    } else {
		status = wind_list(port_data, port_data->output);
	    }
	} while (0);
    }

    return status;
}


static int
roll_data(void *context, void *output, void **input)
{
    int status;
    struct port_type *port_data;

    port_data = context;

    status = port_data->function(port_data->context, output, input);
    if (status) {
    } else {
	int type;
	struct x1f4_dxcast_type *dxcast_data;

	dxcast_data = I_USER(output);

	type = dxcast_data->type;

	do {
	    int link;
#if __LINK_COPY_DATA__
	    int sink;
#endif				/* __LINK_COPY_DATA__ */
#if __LINK_COPY_DATA__
	    struct fxdata_type fxdata;
#endif				/* __LINK_COPY_DATA__ */
	    struct direct_type *direct_data;

	    if (type == X1f4_E4_VOID) {
		status = _libx1f4i0_lxlist_stat_gulf(port_data->lxlist_data);
		break;
	    }

	    direct_data = port_data->cxroll.direct_data;

#if __LINK_COPY_DATA__
	    sink = direct_data->link;
#endif				/* __LINK_COPY_DATA__ */

#if __LINK_COPY_DATA__
	    fxdata = direct_data->fxdata;
#endif				/* __LINK_COPY_DATA__ */

	    status = jack_data
		(port_data, &direct_data->fxdata, dxcast_data, type, &link);

#if __LINK_COPY_DATA__
	    if (~sink) {
		status = free_data(port_data, &fxdata, sink, status);
	    }
#endif				/* __LINK_COPY_DATA__ */

	    port_data->count--;

	    direct_data->link = link;
	    direct_data->type = type;

	    port_data->cxroll.direct_data = direct_data + 1;

	    if (status) {
		break;
	    }
	    if (port_data->count) {
	    } else {
		status = wind_list(port_data, port_data->output);
	    }
	} while (0);
    }

    return status;
}


static int
rule_list(void *line, void *data)
{
    int link, type;
    struct direct_type *direct_data;
    struct fxdata_type fxdata;
    struct rule_type *rule_data;

    rule_data = line;

    direct_data = rule_data->direct_data;

    link = effect_link(data);
    type = effect_type(data);

    type_effect(data, direct_data->type);
    link_effect(data, direct_data->link);

    direct_data->link = link;
    direct_data->type = type;

    fxdata = *(struct fxdata_type *) effect_data(data);

    *(struct fxdata_type *) effect_data(data) = direct_data->fxdata;

    direct_data->fxdata = fxdata;

    rule_data->direct_data++;

    return 0;
}


static int
slip_port(void *port)
{
    int excess, status = 0;
    struct port_type *port_data;
    void *side;

    port_data = port;

    side = port_data->side;
    if (side) {
	struct dxlist_type *dxlist_data;
	unsigned size;

	dxlist_data = port_data->dxlist_data;

	_libx1f4i0_size_l4list(side, &size);
	if (size) {
	    status = _libx1f4i0_lxlist_null_list
		(port_data->lxlist_data, NULL, dxlist_data);

	    _libx1f4i0_swap_l4list(dxlist_data->list, side);

	    dxlist_data->size = port_data->size;
	}

	excess = _libx1f4i0_fini_l4list(&side);
	if (excess) {
	    status = _libx1f4i0_lxlist_stat_list
		(port_data->lxlist_data, status, excess);
	}

	port_data->side = NULL;
    }

    return status;
}


/*
 * must occur before the collection callback setup by _roll_list destroying
 * the temporary array is executed.
 */
static int
spin_port(struct port_type *port_data)
{
    int status;
    unsigned size;

    status = slip_port(port_data);
    if (status) {
    } else {
	size = port_data->size - port_data->count;
	if (size) {
	    struct dxlist_type *dxlist_data;
	    struct rule_type rule;

	    dxlist_data = port_data->dxlist_data;

	    port_data->count = port_data->size;

	    rule.direct_data = port_data->cxroll.direct_data - size;

	    status = _libx1f4i0_over_l4list
		(dxlist_data->list, 0, size - 1, &rule, rule_list);
	}
    }

    return status;
}


static int
swap_port(void *port)
{
    struct dxlist_type *dxlist_data;
    struct port_type *port_data;

    port_data = port;

    dxlist_data = port_data->dxlist_data;

    _libx1f4i0_swap_l4list(dxlist_data->list, port_data->side);

    dxlist_data->size = 0;

#if __LINK_COPY_DATA__
    *port_data->cxroll.miss = port_data->count;
#else
    *port_data->cxroll.miss = 0;
#endif				/* __LINK_COPY_DATA__ */

    return 0;
}


static int
wind_list(struct port_type *port_data, void *output)
{
    int status;

    status = spin_port(port_data);
    if (status) {
    } else {
	l_LIST(port_data->dxlist_data, output);
    }

    return status;
}


int
_libx1f4i0_lxlist_n_pending(void *context, void *output, void **input)
{
    int status;

    do {
	X1f4_E4_C_MODE lock;
	struct dxlist_type *dxlist_data;
	struct fxroll_type fxroll;
	struct lxlist_type *lxlist_data;
	struct x1f4_dxcast_type *dxcast_data;
	const struct x1f4_function_type *function_text;
	struct x1f4_linetext_type *linetext_data;
	unsigned class, count, grade;
	void *beta, **mind, **track, **used;

	if (output == x1f4_a1_walk_link) {
	    track = input[1];
	} else {
	    track = input;
	}

	used = track[-1];

	dxcast_data = I_USER(track[1]);

	if (X1f4_E4_CALL < dxcast_data->type) {
	} else {
	    status = _libx1f4i0_lxlist_stat_roll(context, "l_rform", 7);
	    break;
	}

	linetext_data = I_USER(&dxcast_data->data);

	if (linetext_data->function.function) {
	} else {
	    status = _libx1f4i0_lxlist_stat_zero(context);
	    break;
	}

	if (linetext_data->function.type == X1f4_E4_VOID) {
	    status = _libx1f4i0_lxlist_stat_void(context);
	    break;
	}

	dxlist_data = I_LIST(track[0]);

	lock = I_MODE(track[2]);

	function_text = used[X1f4_E4_RENDITION];

	class = function_text->count - 3;

	grade = class + 1;

	if (lock < -(int) grade || !(lock < (int) grade)) {
	    status = _libx1f4i0_lxlist_stat_land(context, grade, lock);

	    break;
	}

	if (lock < (X1f4_E4_C_MODE) 0) {
	    lock += grade;
	}

	count = dxlist_data->size;
	if (count) {
	    if (output == x1f4_a1_walk_link) {
		if (linetext_data->function.flags
		    & (X1f4_E4_LEAD_CALL | X1f4_E4_WALK_LINK)) {
		} else {
		    input[2] = NULL;
		    output = input[0];
		    input = input[1];
		}
	    }

	} else {
	    status = 0;

	    if (output == x1f4_a1_walk_link) {
		input[2] = NULL;
		output = input[0];
	    }

	    l_LIST(dxlist_data, output);

	    break;
	}

	lxlist_data = context;

	status = lxlist_data->link_v.pick
	    (lxlist_data->link_v.data, &mind, slip_port,
	     sizeof(struct port_type), &beta);
	if (status) {
	    status = _libx1f4i0_lxlist_stat_link(lxlist_data);
	} else {
	    int type;
	    struct port_type *port_data;
	    struct x1f4_linetext_type linetext;
	    struct x1f4_linetext_type *linetext_roll;

	    *mind = beta;

	    port_data = beta;

	    status = _libx1f4i0_lxlist_long_list
		(lxlist_data, &port_data->side);
	    if (status) {
		port_data->side = NULL;
		break;
	    } else {
	    }

	    port_data->context = (void *) linetext_data->context;
	    port_data->function = linetext_data->function.function;

	    port_data->count = count;

	    port_data->dxlist_data = dxlist_data;
	    port_data->lxlist_data = lxlist_data;

	    if (output == x1f4_a1_walk_link) {
		linetext_roll = linetext_data;
	    } else {
		port_data->output = output;

		linetext_roll = &linetext;

		linetext.function = linetext_data->function;

		linetext.context = port_data;
		linetext.function.function = roll_data;
	    }

	    port_data->excase_data = lock_excase(dxlist_data);
	    port_data->lxtext_data = lxlist(context)->link_f.text;

	    type = linetext_data->function.type;

	    if (type != X1f4_E4_CASE) {
		struct lxtext_type *lxtext_data;
		unsigned miss;

		port_data->type = type;

		lxtext_data = port_data->lxtext_data;

		close_frame(miss, lxtext_data);

		linetext.function.function = poll_data;

		if (miss) {
		    port_data->link = lxtext_data - port_data->lxtext_data;
		} else {
		    port_data->link = ~0;
		}

		port_data->lxtext_lift = lxtext_data;

		if (lxtext_data->lxtype.flags & X1f4_LX_LINK_ACCESS) {
		    port_data->jack = fast_data;
		} else {
		    if (lxtext_data->lxtype.copy) {
			port_data->jack = copy_data;
		    } else {
			port_data->jack = ever_data;
		    }
		}
	    }

	    port_data->size = count;

#if __LINK_COPY_DATA__
# define Copy				1
#else
# define Copy				0
#endif				/* __LINK_COPY_DATA__ */

	    fxroll.copy = Copy;
	    fxroll.head = 0;
	    fxroll.tail = count;

	    fxroll.infer = 0;

	    status = _libx1f4i0_lxrail_unto_rail
		(context, output, input, track, linetext_roll, count,
		 function_text, 3, lock, dxlist_data, &port_data->cxroll,
		 port_data, swap_port, &fxroll, _libx1f4i0_lxlist_ever_list);
	    if (status) {
		if (output == x1f4_a1_walk_link) {
		} else {
		    spin_port(port_data);
		}
	    } else {
		if (output == x1f4_a1_walk_link) {
		    struct x1f4_dispatch_type *dispatch_data;

		    dispatch_data = input[2];

		    port_data->back = dispatch_data->back;
		    port_data->call = dispatch_data->call;

		    dispatch_data->back = port_data;
		    dispatch_data->call =
			type != X1f4_E4_CASE ? less_data : miss_data;

		    port_data->dispatch_data = dispatch_data;
		}
	    }
	}
    } while (0);

    return status;
}
